@@ -26,10 +26,12 @@
 // request the std format macros
 #define __STDC_FORMAT_MACROS
 
+#define PY_SSIZE_T_CLEAN
 #include <Python.h>
 #include <stdlib.h>
 #include <inttypes.h>
 #include <libcec/cec.h>
+#include <algorithm>
 #include <list>
 
 #include "device.h"
@@ -103,7 +105,7 @@ using namespace CEC;
 #define HAVE_CEC_ADAPTER_DESCRIPTOR 0
 #endif
 
-int parse_physical_addr(char * addr) {
+int parse_physical_addr(const char * addr) {
    int a, b, c, d;
    if( sscanf(addr, "%x.%x.%x.%x", &a, &b, &c, &d) == 4 ) {
       if( a > 0xF || b > 0xF || c > 0xF || d > 0xF ) return -1;
@@ -156,7 +158,7 @@ std::list<CEC_ADAPTER_TYPE> get_adapters() {
       dev_list = (CEC_ADAPTER_TYPE*)realloc(dev_list, 
          cec_count * sizeof(CEC_ADAPTER_TYPE));
       count = CEC_adapter->CEC_FIND_ADAPTERS(dev_list, cec_count);
-      count = std::min(count, cec_count);
+      count = (std::min)(count, cec_count);
    }
 
    for( int i=0; i<count; i++ ) {
@@ -286,16 +288,12 @@ static PyObject * list_devices(PyObject * self, PyObject * args) {
    return result;
 }
 
-class Callback {
+struct Callback {
    public:
       long int event;
       PyObject * cb;
 
       Callback(long int e, PyObject * c) : event(e), cb(c) {
-         Py_INCREF(cb);
-      }
-      ~Callback() {
-         Py_DECREF(cb);
       }
 };
 
@@ -319,6 +317,7 @@ static PyObject * add_callback(PyObject * self, PyObject * args) {
          return NULL;
       }
 
+      Py_INCREF(callback);
       Callback new_cb(events, callback);
 
       debug("Adding callback for event %ld\n", events);
@@ -332,7 +331,7 @@ static PyObject * add_callback(PyObject * self, PyObject * args) {
 
 static PyObject * remove_callback(PyObject * self, PyObject * args) {
   PyObject * callback;
-  long int events = EVENT_ALL; // default to all events
+  Py_ssize_t events = EVENT_ALL; // default to all events
 
   if( PyArg_ParseTuple(args, "O|i:remove_callback", &callback, &events) ) {
      for( cb_list::iterator itr = callbacks.begin(); 
@@ -344,12 +343,38 @@ static PyObject * remove_callback(PyObject * self, PyObject * args) {
            if( itr->event == 0 ) {
               // if this callback has no events, remove it
               itr = callbacks.erase(itr);
+              Py_DECREF(callback);
            }
         }
      }
      
   }
-  return NULL;
+  Py_INCREF(Py_None);
+  return Py_None;
+}
+
+static PyObject * make_bound_method_args(PyObject * self, PyObject * args) {
+   Py_ssize_t count = 0;
+   if( PyTuple_Check(args) ) {
+      count = PyTuple_Size(args);
+   }
+   PyObject * result = PyTuple_New(count+1);
+   if( result == NULL ) {
+      return NULL;
+   }
+   assert(self != NULL);
+   Py_INCREF(self);
+   PyTuple_SetItem(result, 0, self);
+   for( Py_ssize_t i=0; i<count; i++ ) {
+      PyObject * arg = PyTuple_GetItem(args, i);
+      if( arg == NULL ) {
+         Py_DECREF(result);
+         return NULL;
+      }
+      Py_INCREF(arg);
+      PyTuple_SetItem(result, i+1, arg);
+   }
+   return result;
 }
 
 static PyObject * trigger_event(long int event, PyObject * args) {
@@ -366,8 +391,21 @@ static PyObject * trigger_event(long int event, PyObject * args) {
       //debug("Checking callback %d with events %ld\n", i, itr->event);
       if( itr->event & event ) {
          //debug("Calling callback %d\n", i);
+         PyObject * callback = itr->cb;
+         PyObject * arguments = args;
+         if( PyMethod_Check(itr->cb) ) {
+            callback = PyMethod_Function(itr->cb);
+            PyObject * self = PyMethod_Self(itr->cb);
+            if( self ) {
+               // bound method, prepend self/cls to argument tuple
+               arguments = make_bound_method_args(self, args);
+            }
+         }
          // see also: PyObject_CallFunction(...) which can take C args
-         PyObject * temp = PyObject_CallObject(itr->cb, args);
+         PyObject * temp = PyObject_CallObject(callback, arguments);
+         if( arguments != args ) {
+            Py_XDECREF(arguments);
+         }
          if( temp ) {
             debug("Callback succeeded\n");
             Py_DECREF(temp);
@@ -384,26 +422,42 @@ static PyObject * trigger_event(long int event, PyObject * args) {
 }
 
 static PyObject * transmit(PyObject * self, PyObject * args) {
+   unsigned char initiator = 'g';
    unsigned char destination;
    unsigned char opcode;
    const char * params = NULL;
-   int param_count = 0;
+   Py_ssize_t param_count = 0;
 
-   if( PyArg_ParseTuple(args, "bb|s#:transmit", &destination, &opcode,
-         &params, &param_count) ) {
+   if( PyArg_ParseTuple(args, "bb|s#b:transmit", &destination, &opcode,
+         &params, &param_count, &initiator) ) {
       if( destination < 0 || destination > 15 ) {
          PyErr_SetString(PyExc_ValueError, "Logical address must be between 0 and 15");
          return NULL;
       }
+      if( initiator != 'g' ) {
+         if( initiator < 0 || initiator > 15 ) {
+            PyErr_SetString(PyExc_ValueError, "Logical address must be between 0 and 15");
+            return NULL;
+         }
+      } else {
+         initiator = CEC_adapter->GetLogicalAddresses().primary;
+      }
+      if( param_count > CEC_MAX_DATA_PACKET_SIZE ) {
+         char errstr[1024];
+         snprintf(errstr, 1024, "Too many parameters, maximum is %d",
+            CEC_MAX_DATA_PACKET_SIZE);
+         PyErr_SetString(PyExc_ValueError, errstr);
+         return NULL;
+      }
       cec_command data;
       bool success;
       Py_BEGIN_ALLOW_THREADS
-      data.initiator = CEC_adapter->GetLogicalAddresses().primary;
+      data.initiator = (cec_logical_address)initiator;
       data.destination = (cec_logical_address)destination;
       data.opcode = (cec_opcode)opcode;
       data.opcode_set = 1;
       if( params ) {
-         for( int i=0; i<param_count; i++ ) {
+         for( Py_ssize_t i=0; i<param_count; i++ ) {
             data.PushBack(((uint8_t *)params)[i]);
          }
       }
@@ -484,7 +538,7 @@ static PyObject * set_stream_path(PyObject * self, PyObject * args) {
          }
 #if PY_MAJOR_VERSION >= 3
       } else if(PyUnicode_Check(arg)) {
-         char * arg_s = PyUnicode_AsUTF8(arg);
+         const char * arg_s = PyUnicode_AsUTF8(arg);
 #else
       } else if(PyString_Check(arg)) {
          char * arg_s = PyString_AsString(arg);
@@ -568,14 +622,22 @@ PyObject * set_port(PyObject * self, PyObject * args) {
 
 PyObject * can_persist_config(PyObject * self, PyObject * args) {
    if( PyArg_ParseTuple(args, ":can_persist_config") ) {
+#if CEC_LIB_VERSION_MAJOR >= 5
+      RETURN_BOOL(CEC_adapter->CanSaveConfiguration());
+#else
       RETURN_BOOL(CEC_adapter->CanPersistConfiguration());
+#endif
    }
    return NULL;
 }
 
 PyObject * persist_config(PyObject * self, PyObject * args) {
    if( PyArg_ParseTuple(args, ":persist_config") ) {
+#if CEC_LIB_VERSION_MAJOR >= 5
+      if( ! CEC_adapter->CanSaveConfiguration() ) {
+#else
       if( ! CEC_adapter->CanPersistConfiguration() ) {
+#endif
          PyErr_SetString(PyExc_NotImplementedError,
                "Cannot persist configuration");
          return NULL;
@@ -585,7 +647,11 @@ PyObject * persist_config(PyObject * self, PyObject * args) {
          PyErr_SetString(PyExc_IOError, "Could not get configuration");
          return NULL;
       }
+#if CEC_LIB_VERSION_MAJOR >= 5
+      RETURN_BOOL(CEC_adapter->SetConfiguration(&config));
+#else
       RETURN_BOOL(CEC_adapter->PersistConfiguration(&config));
+#endif
    }
    return NULL;
 }
@@ -635,12 +701,17 @@ int log_cb(void * self, const cec_log_message message) {
    long int time = message.time;
    const char* msg = message.message;
 #endif
-   PyObject * args = Py_BuildValue("(iils)", EVENT_LOG, 
+   // decode message ignoring invalid characters
+   PyObject * umsg = PyUnicode_DecodeASCII(msg, strlen(msg), "ignore");
+   PyObject * args = Py_BuildValue("(iilO)", EVENT_LOG,
          level,
          time,
-         msg);
-   trigger_event(EVENT_LOG, args);
-   Py_DECREF(args);
+         umsg);
+   if( args ) {
+      trigger_event(EVENT_LOG, args);
+      Py_DECREF(args);
+   }
+   Py_XDECREF(umsg);
    PyGILState_Release(gstate);
 #if CEC_LIB_VERSION_MAJOR >= 4
    return;
@@ -667,8 +738,10 @@ int keypress_cb(void * self, const cec_keypress key) {
    PyObject * args = Py_BuildValue("(iBI)", EVENT_KEYPRESS,
          keycode,
          duration);
-   trigger_event(EVENT_KEYPRESS, args);
-   Py_DECREF(args);
+   if( args ) {
+      trigger_event(EVENT_KEYPRESS, args);
+      Py_DECREF(args);
+   }
    PyGILState_Release(gstate);
 #if CEC_LIB_VERSION_MAJOR >= 4
    return;
@@ -677,6 +750,22 @@ int keypress_cb(void * self, const cec_keypress key) {
 #endif
 }
 
+static PyObject * convert_cmd(const cec_command* cmd) {
+#if PY_MAJOR_VERSION >= 3
+   return Py_BuildValue("{sBsBsOsOsBsy#sOsi}",
+#else
+   return Py_BuildValue("{sBsBsOsOsBss#sOsi}",
+#endif
+         "initiator", cmd->initiator,
+         "destination", cmd->destination,
+         "ack", cmd->ack ? Py_True : Py_False,
+         "eom", cmd->eom ? Py_True : Py_False,
+         "opcode", cmd->opcode,
+         "parameters", cmd->parameters.data, cmd->parameters.size,
+         "opcode_set", cmd->opcode_set ? Py_True : Py_False,
+         "transmit_timeout", cmd->transmit_timeout);
+}
+
 #if CEC_LIB_VERSION_MAJOR >= 4
 void command_cb(void * self, const cec_command* command) {
 #else
@@ -685,12 +774,16 @@ int command_cb(void * self, const cec_command command) {
    debug("got command callback\n");
    PyGILState_STATE gstate;
    gstate = PyGILState_Ensure();
-   // TODO: figure out how to pass these parameters
-   //  we'll probably have to build an Object for this
-   PyObject * args = Py_BuildValue("(i)", EVENT_COMMAND);
-   // don't bother triggering an event until we can actually pass arguments
-   //trigger_event(EVENT_COMMAND, args);
-   Py_DECREF(args);
+#if CEC_LIB_VERSION_MAJOR >= 4
+   const cec_command * cmd = command;
+#else
+   const cec_command * cmd = &command;
+#endif
+   PyObject * args = Py_BuildValue("(iO&)", EVENT_COMMAND, convert_cmd, cmd);
+   if( args ) {
+      trigger_event(EVENT_COMMAND, args);
+      Py_DECREF(args);
+   }
    PyGILState_Release(gstate);
 #if CEC_LIB_VERSION_MAJOR >= 4
    return;
@@ -714,9 +807,11 @@ int config_cb(void * self, const libcec_configuration) {
    //  this will probably be _lots_ of work and should probably wait until
    //  a later release, or when it becomes necessary.
    PyObject * args = Py_BuildValue("(i)", EVENT_CONFIG_CHANGE);
-   // don't bother triggering an event until we can actually pass arguments
-   //trigger_event(EVENT_CONFIG_CHANGE, args);
-   Py_DECREF(args);
+   if( args ) {
+      // don't bother triggering an event until we can actually pass arguments
+      //trigger_event(EVENT_CONFIG_CHANGE, args);
+      Py_DECREF(args);
+   }
    PyGILState_Release(gstate);
 #if CEC_LIB_VERSION_MAJOR >= 4
    return;
@@ -740,8 +835,10 @@ int alert_cb(void * self, const libcec_alert alert, const libcec_parameter p) {
       Py_INCREF(param);
    }
    PyObject * args = Py_BuildValue("(iiN)", EVENT_ALERT, alert, param);
-   trigger_event(EVENT_ALERT, args);
-   Py_DECREF(args);
+   if( args ) {
+      trigger_event(EVENT_ALERT, args);
+      Py_DECREF(args);
+   }
    PyGILState_Release(gstate);
 #if CEC_LIB_VERSION_MAJOR >= 4
    return;
@@ -755,8 +852,10 @@ int menu_cb(void * self, const cec_menu_state menu) {
    PyGILState_STATE gstate;
    gstate = PyGILState_Ensure();
    PyObject * args = Py_BuildValue("(ii)", EVENT_MENU_CHANGED, menu);
-   trigger_event(EVENT_MENU_CHANGED, args);
-   Py_DECREF(args);
+   if( args ) {
+      trigger_event(EVENT_MENU_CHANGED, args);
+      Py_DECREF(args);
+   }
    PyGILState_Release(gstate);
    return 1;
 }
@@ -769,8 +868,10 @@ void activated_cb(void * self, const cec_logical_address logical_address,
    PyObject * active = (state == 1) ? Py_True : Py_False;
    PyObject * args = Py_BuildValue("(iOi)", EVENT_ACTIVATED, active,
       logical_address);
-   trigger_event(EVENT_ACTIVATED, args);
-   Py_DECREF(args);
+   if( args ) {
+      trigger_event(EVENT_ACTIVATED, args);
+      Py_DECREF(args);
+   }
    PyGILState_Release(gstate);
    return;
 }
