@@ -22,6 +22,7 @@
  * Author: Austin Hendrix <namniart@gmail.com>
  */
 
+#define PY_SSIZE_T_CLEAN
 #include <Python.h>
 
 #include <libcec/cec.h>
@@ -39,3 +40,10 @@ struct Device {
 };
 
 PyTypeObject * DeviceTypeInit(CEC::ICECAdapter * adapter);
+
+/*
+ * Compat for libcec 3.x
+ */
+#if CEC_LIB_VERSION_MAJOR < 4
+  #define CEC_MAX_DATA_PACKET_SIZE (16 * 4)
+#endif
